/*
Victor Ng
vn6216@albany.edu
csi402 Spring 2006

*/


#include <stdio.h>
#include <stdlib.h>

#include "aconstants.h"
  #include "astruct_def.h"
#include "aexterns.h"

void readcfgfile( FILE * finp )
{
  int i=0; // counter
  char *line; // pointer to string
  
  fscanf( finp, "%d", &nfiles); // read the number of files to be expected
  fscanf( finp, "%d", &startaddr); // read the starting address of the executible
  
  while (i < nfiles)
  { // while the number of files read != number of files to be expected
    if ( (line = (char *) malloc(MAXLEN * sizeof(char)) ) == NULL ) // allocating memory of 'MAXLEN' characters
    { // error in allocating memory 
      fprintf(stderr, "Err: cannot allocate memory. \n"); // print error message 
      exit(1); // exit due to insufficient memory 
    }; // memory allocation successful
    
    fscanf( finp, "%s", line); // read the name from the config file
    fname[i]=line; // save the name into the array
    i++;// increment counter i
  };// end of read
  
}

void writetextbits(long int textbits, unsigned int addr)
{
  objout[addr] = textbits; // write to objout
}

void initobj ()
{
  int i=0; // counter
  while (i <= MAXMEM)
  { // i <= MAXMEM
    writetextbits(0, i); // write zero to addr i
    i++; // increment counter
  };// end of initialize
}

int searchnodes( char * nameread )
{
  pNode pPointer = pHeader; // pointer to current node, start from the head of linked list
  
  while (pPointer != NULL)
  { // while there are still nodes in the list
    if (strcmp(nameread, pPointer->name) == 0) // compare the string with the name of current node
      return 0; // if the name matches current node, return 0;
    
    // name doesn't match
    if (pPointer->next != NULL) // check if next node is NULL
      pPointer = pPointer->next; // next node is not NULL, increment pointer
    else  // next node is NULL
      pPointer = NULL; // pointer is NULL;
  }; // end of search
  // at this point, the string doesn't match any in the nodes
  return 1; // return 1
}

unsigned int findaddr( char * nameread )
{
  pNode pPointer = pHeader; // pointer to current node, start from the head of linked list
  
  while (pPointer != NULL)
  { // while there are still nodes in the list
    if (strcmp(nameread, pPointer->name) == 0) // compare the string with the name of current node
      return pPointer->absaddr; // if the name matches current node, return the address;
    
    // name doesn't match
    if (pPointer->next != NULL) // check if next node is NULL
      pPointer = pPointer->next; // next node is not NULL, increment pointer
    else  // next node is NULL
      pPointer = NULL; // pointer is NULL;
  }; // end of search
  // at this point, the string doesn't match any in the nodes
  fprintf(stderr, "Err: Undefined symbol '%s'.\n", nameread); // prints error 
  exit(1); // exit due to wrong header record 
  
}

void insertnode (char * nameread, unsigned int addr)
{
  pNode newNode; // pointer to newly created note, created useing malloc
  
  // check for multi defined symbol
  if (searchnodes( nameread ) == 0)
  { // symbol has been defined
    fprintf(stderr, "Err: multiply-defined symbol '%s'.\n", nameread); // prints error 
    exit(1); // exit due to wrong header record 
  } // symbol has not been defined
  newNode = malloc( sizeof( Node )); // allocate new node space
  newNode->next = NULL; // setting NULL to next pointer
  strcpy(newNode->name, nameread); // saving name to node
  newNode->absaddr = addr; // saving address to node
  if (pHeader == NULL) // if pHeader is NULL
    pHeader = newNode; // no nodes exist, this is first node
  else // nodes already exist
    pCurrent->next = newNode; // put newNode at end of linked list
  pCurrent = newNode; // increment new pointer.
}

void checkaddr (char * fname, unsigned int addr)
{
  if (! ((addr >= 0) && (addr <= MAXMEM)) ) //if address is not between 0 and MAXMEM
  { // address is greater than MAXMEM
    fprintf(stderr, "Err: address out of range in '%s'. \n", fname); // prints error 
    exit(1); // exit due to addressing error
  };
}

void modobj( unsigned int targetaddr, unsigned int foundaddr)
{
  unsigned long int objtext; // temp store to text bits that are read
  unsigned long int targetshifted=0; // tempstore for the shift

  objtext = objout[targetaddr]; // read the text bits
  targetshifted += foundaddr; // conver the offset type
//  targetshifted = foundaddr >> INTSHIFT; // shift the bits
  objtext += targetshifted; // adds the offset
  
  objout[targetaddr] = objtext; // write to objout
}

void readobjects ( FILE * foutp )
{
  // pointers for use with strcmp
  static const char * cHdr = "H"; // H for header record
  static const char * cDef = "D"; // D for Def. record
  static const char * cTxt = "T"; // T for Text record
  static const char * cMod = "M"; // M for Modification record
  
  FILE * finp; // file input pointer
  char recordtype; // temp store for recordtype
  char nameread[NAMELEN]; // tempstore for label name
  unsigned int addrread; // tempstore for address read from file
  unsigned long int textbits; // tempstore for textbits from T-Record
  
  unsigned int curaddr, modstartaddr; // temp variables
  unsigned int foundaddr, targetaddr; // temp variables
  
  int i=0; // counters
  
//First Pass 
  curaddr = startaddr; // set to starting address
  i=0; // reset counter
  while ( i < nfiles)
  { // while there are still files to process
    modstartaddr = curaddr; // saving module starting address
    if ( (finp = fopen(fname[i], "r")) == NULL )  // use specified name
    {   // open failed 
      fprintf(stderr, "Could not open file '%s' for reading. \n", fname[i]); // prints error 
      exit(1); // exit due to failure in opening the file 
    };// open successful
    
    fscanf( finp, "%s", &recordtype); // read record type name 
    if (strcmp(&recordtype, cHdr) != 0)
    { // double check for H record
      fprintf(stderr, "Err: wrong header record in file '%s'.\n", fname[i]); // prints error 
      exit(1); // exit due to wrong header record 
    } // H-Record
    
    // process the header record
    fscanf( finp, "%s", nameread); // read module name 
    
    checkaddr (fname[i], curaddr); // check address
    insertnode(nameread, curaddr); // insert into symbol table
    
    while (!feof(finp))
    { // while not end of file
      fscanf( finp, "%s", &recordtype); // read record type name 
      if (feof(finp))
      { // if end of file
        ; // do nothing, skip to next file
      }
      else if (strcmp(&recordtype, cDef) == 0)
      { // D-Record
        fscanf( finp, "%s", nameread); // read module name 
        fscanf( finp, "%d", &addrread); // read relative address
        addrread += curaddr; // adding relative address
        checkaddr (fname[i], addrread); // check address
        insertnode(nameread, addrread); // insert into symbol table        
      }
      else if (strcmp(&recordtype, cTxt) == 0)
      { // T-Record
        fscanf( finp, "%i", &textbits); // read the textbits
        curaddr++; // increment the address counter
      }
      else if (strcmp(&recordtype, cMod) == 0)
      { // M-Record
        fscanf( finp, "%d", &addrread); // read the number
        fscanf( finp, "%s", nameread); // read the name
      }
      else 
      { // syntax error (should not happen)
        fprintf(stderr, "Err: syntax error - unknown record type:'%s'. \n", &recordtype); // prints error 
        exit(1); // exit due to syntax error
      };
      
    }; // eof reached
    // address checking
    checkaddr (fname[i], addrread); // check address
    fclose(finp); // close finp
    
    i++; // increment counter
  }; // end of first past
  
//Second Pass 
  curaddr = startaddr; // set to starting address
  i=0; // reset counter
  while ( i < nfiles )
  { // while there are still files to process
    modstartaddr = curaddr; // saving module starting address
    if ( (finp = fopen(fname[i], "r")) == NULL )  // use specified name
    {   // open failed 
      fprintf(stderr, "Could not open file '%s' for reading. \n", fname[i]); // prints error 
      exit(1); // exit due to failure in opening the file 
    };// open successful
    fscanf( finp, "%s", &recordtype); // read record type name 
    if (strcmp(&recordtype, cHdr) != 0)
    { // double check for H record
      fprintf(stderr, "Err: wrong header record in file '%s'.\n", fname[i]); // prints error 
      exit(1); // exit due to wrong header record 
    } // H-Record
    fscanf( finp, "%s", nameread); // read module name 

    while (!feof(finp))
    { // while not end of file
      fscanf( finp, "%s", &recordtype); // read record type name 
      if (feof(finp))
      { // if end of file
        ; // do nothing, skip to next file
      }
      else if (strcmp(&recordtype, cDef) == 0)
      { // D-Record
        fscanf( finp, "%s", nameread); // read module name 
        fscanf( finp, "%d", &addrread); // read relative address     
      }
      else if (strcmp(&recordtype, cTxt) == 0)
      { // T-Record
        fscanf( finp, "%i", &textbits); // read the textbits
        writetextbits(textbits, curaddr); // write the textbits into mem
        curaddr++; // increment the address counter
      }
      else if (strcmp(&recordtype, cMod) == 0)
      { // M-Record
        fscanf( finp, "%d", &addrread); // read the number
        fscanf( finp, "%s", nameread); // read the name
        foundaddr = findaddr( nameread ); // find the address from symbol table
        targetaddr = modstartaddr + addrread; // target address that needs to be modified
        modobj(targetaddr, foundaddr); // modify object[]
      }
      else 
      { // syntax error (should not happen)
        fprintf(stderr, "Err: syntax error - unknown record type:'%s'. \n", &recordtype); // prints error 
        exit(1); // exit due to syntax error
      };
      
    }; // eof reached
    // address checking
    checkaddr (fname[i], curaddr); // check address range
    fclose(finp); // close finp
    
    i++; // increment counter
  }; // end of second past
  
// now to write the module out
  modstartaddr = curaddr - startaddr; // find the size
  targetaddr = modstartaddr << BITSHIFT; // shift the size left so it's 2 bytes
  foundaddr = startaddr << BITSHIFT; // // shift the starting address left so it's 2 bytes
  
  fwrite ( &targetaddr, 2,1, foutp ); // write 2 bytes of shifted size into foutp 
  fwrite ( &foundaddr, 2,1, foutp ); // write 2 bytes of shifted addr into foutp
  i=startaddr; // reset counter
  while (i < curaddr)
  { // while there are still T-records to be written
    fwrite ( &objout[i], sizeof(long int),1, foutp ); // write to objout in binary
    i++; // increment counter
  };
  
}




